package com.hammurapi.review;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;

import com.hammurapi.review.Inspector;
import com.hammurapi.review.Measurement;
import com.hammurapi.review.Observation;
import com.hammurapi.review.Violation;
import com.hammurapi.review.Warning;

/**
 * Aggregates violations by severity and inspector and measurements 
 * by name into metrics.
 * @author Pavel Vlasov
 *
 */
public class Aggregator implements ObservationSink {
	
	public Aggregator() {
		// Default constructor.
	}
	
	/**
	 * Aggregates findings reported by the given language element and its children.
	 * @param languageElement
	 * @throws ReviewException 
	 */
	public Aggregator(LanguageElement languageElement) throws ReviewException {
		for (Observation o: languageElement.getObservations()) {
			consumeObservation(o);
		}
		TreeIterator<EObject> it = languageElement.eAllContents();
		while (it.hasNext()) {
			EObject next = it.next();
			if (next instanceof LanguageElement) {
				for (Observation o: ((LanguageElement) next).getObservations()) {
					consumeObservation(o);
				}				
			}
		}
	}

	@Override
	public void close() throws ReviewException {
		// NOP
		
	}
	
	private List<Warning> warnings = new ArrayList<Warning>();
	
	public static class Metric {
		private String name;
		
		public Metric(String name) {
			this.name = name;
		}
		
		public String getName() {
			return name;
		}
		
		private double min;
		private double max;
		private int total;
		private List<Measurement> measurements = new ArrayList<Measurement>();
		
		synchronized void addMeasurement(Measurement measurement) {
			if (measurements.isEmpty() || measurement.getValue()<min) {
				min = measurement.getValue();					
			}
			if (measurements.isEmpty() || measurement.getValue()>max) {
				max = measurement.getValue();					
			}
			measurements.add(measurement);
			total+=measurement.getValue();
		}

		public List<Measurement> getMeasurements() {
			Collections.sort(measurements, new Comparator<Measurement>()  {

				@Override
				public int compare(Measurement o1, Measurement o2) {
					if (o1==o2) {
						return 0;
					}
					if (o1.getValue()<o2.getValue()) {
						return -1;
					}
					if (o1.getValue()>o2.getValue()) {
						return 1;
					}
					return o1.hashCode()-o2.hashCode();
				}
			});
			return measurements;
		}
		
		public double getMax() {
			return max;
		}
		
		public double getMin() {
			return min;
		}
		
		public int getTotal() {
			return total;
		}
		
		public double getAverage() {
			return total/measurements.size();
		}
	}
	
	private Map<Integer, Severity> severities = new TreeMap<Integer, Severity>();
	
	public static class Severity {
		Integer severity;
		
		public Severity(Integer severity) {
			this.severity = severity;
		}
		
		Map<String, InspectorEntry> inspectors = new TreeMap<String, InspectorEntry>();
		
		public Collection<InspectorEntry> getInspectors() {
			return inspectors.values();
		}
		
		public Integer getSeverity() {
			return severity;
		}
	}
	
	public static class InspectorEntry {
		List<Violation> violations = new ArrayList<Violation>();
		Inspector inspector;
		
		public Inspector getInspector() {
			return inspector;
		}
		
		public List<Violation> getViolations() {
			return violations;
		}
	}
	
	private Map<String, Metric> metrics = new TreeMap<String, Metric>();

	@Override
	public synchronized void consumeObservation(Observation observation) throws ReviewException {
		if (observation instanceof Violation) {
			Violation violation = (Violation) observation;
			Inspector reportedBy = violation.getReportedBy();
			if (reportedBy!=null) {
				Severity severity = severities.get(reportedBy.getSeverity());
				if (severity==null) {
					severity = new Severity(reportedBy.getSeverity());
					severities.put(reportedBy.getSeverity(), severity);
				}
				InspectorEntry ie = severity.inspectors.get(reportedBy.getName());
				if (ie==null) {
					ie = new InspectorEntry();
					ie.inspector = reportedBy;
					severity.inspectors.put(reportedBy.getName(), ie);
				}
				ie.violations.add(violation);
			}
		} else if (observation instanceof Warning) {
			warnings.add((Warning) observation);
		} else if (observation instanceof Measurement) {
			Measurement measurement = (Measurement) observation;
			Metric metric = metrics.get(measurement.getName());
			if (metric==null) {
				metric = new Metric(measurement.getName());
				metrics.put(measurement.getName(), metric);
			}
			metric.addMeasurement(measurement);
		}		
	}

	public Collection<Severity> getSeverities() {
		return severities.values();
	}
	
	public Collection<Metric> getMetrics() {
		return metrics.values();
	}

	public List<Warning> getWarnings() {
		return warnings;
	}
		
}
