/**
 * Property of Hammurapi Group
 */
package com.hammurapi.review;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * This annotation indicates that method is an inspector method.
 * 
 * Inspector methods can return facts and post facts (conclusions). Review engine
 * needs to know types of rule outputs. From rule class introspection
 * the rule system knows about rule method return type, but it doesn't know
 * about types posted through <code>post()</code>.
 * 
 * Also, return type might not be enough, as returned instances may implement
 * interfaces which other rules are interested in, but which are not declared
 * in the rule return type.
 * 
 * Annotation parameter allows rules to inform the inference system about posted types.
 * <P>
 * For methods without fact types information provided through annotation parameter, 
 * the review system uses method return type. 
 * 
 * @author Pavel
 *
 */
@Target(ElementType.METHOD)
@Retention(RetentionPolicy.RUNTIME)
public @interface Inspect {
	
	/**
	 * @return Conclusion types posted by this method.
	 */
	Class<?>[] value() default {};
	
	/**
	 * @return Inspector severity. This value is overriden by the value provided in the inspector set definition.
	 */
	int severity() default 2;
	
	/**
	 * @return Inspector category. Category definition in the inspector set definition file overrides this category definition.
	 */
	String[] category() default {};
	
}
