package com.hammurapi.review;

import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;
import java.util.logging.Logger;

import org.eclipse.emf.common.util.TreeIterator;

import com.hammurapi.config.Factory;
import com.hammurapi.config.Path;
import com.hammurapi.config.bootstrap.ConfigurationException;
import com.hammurapi.config.bootstrap.Destroyable;
import com.hammurapi.config.bootstrap.FactoryClosure;
import com.hammurapi.config.runtime.CompositeDestroyable;
import com.hammurapi.config.runtime.ConfigurationContext;
import com.hammurapi.config.runtime.DestroyableSink;
import com.hammurapi.config.runtime.FactoryConfig;
import com.hammurapi.convert.ConvertingService;
import com.hammurapi.reasoning.ExceptionHandler;
import com.hammurapi.reasoning.ForwardReasoningSession;
import com.hammurapi.reasoning.ForwardReasoningSessionFactory;
import com.hammurapi.reasoning.ReasoningException;
import com.hammurapi.util.InlineExecutor;
import com.hammurapi.util.SingletonChainingContext;
import com.hammurapi.util.concurrent.NonBlockingFutureTask;

/**
 * This class performs reviews invoking other review framework components.
 * @author Pavel Vlasov
 *
 */
public abstract class ReviewDriverBase implements Callable<Object>, com.hammurapi.config.runtime.Component<Component> {
	
	private static final Logger logger = Logger.getLogger(ReviewDriverBase.class.getName());
	
	private Executor executor;
	private ConfigurationContext<Component> context;
	
	@Override
	public void init(ConfigurationContext<Component> context) throws ConfigurationException {
		executor = context.lookup(Executor.class);
		if (executor==null) {
			executor = InlineExecutor.INSTANCE;
		}
		this.context = context;
	}
	
	/**
	 * Loads component definition resource and executes review.
	 * @param componentConfig
	 * @throws ConfigurationException
	 */
	@SuppressWarnings("unchecked")
	@Override
	public Object call() throws Exception {		
		CompositeDestroyable destroyableSink = new CompositeDestroyable();
		try {
			// Create reporters
			List<Reporter<Object>> reporters = new ArrayList<Reporter<Object>>(); 
			for (Factory rf: context.getDefinition().getReporter()) {
				FactoryClosure<Object> rfr = rf.create(context.getFactoryConfig());
				logger.fine("Creating a reporter");
				reporters.add((Reporter<Object>) rfr.create());
				destroyableSink.addDestroyable(rfr.getDestroyable());
			}
			
			try {
				List<Future<Object>> tasks = new ArrayList<Future<Object>>();
				// Review modules 
				for (Module module: context.getDefinition().getModule()) {
					NonBlockingFutureTask<Object> task = new NonBlockingFutureTask<Object>(createModuleTask(module, reporters, context.getFactoryConfig(), destroyableSink, executor));
					tasks.add(task);
					executor.execute(task);
				}
				
				for (Future<Object> task: tasks) {
					try {
						task.get();
					} catch (Exception e) {
						for (Reporter r: reporters) {
							r.onException(e);
						}
					}
				}
			} finally {
				for (Reporter<Object> r: reporters) {
					r.close();
				}
			}										
		} finally {
			destroyableSink.destroy();
		}
		return null;
	}

	private Callable<Object> createModuleTask(
			final Module module, 
			final List<Reporter<Object>> reporters, 
			final FactoryConfig factoryConfig, 
			final DestroyableSink destroyableSink, 
			final Executor executor) {
		
		logger.fine("Creating a module task for "+module.getName());
		
		return new Callable<Object>() {

			@SuppressWarnings("unchecked")
			@Override
			public Object call() throws Exception {
				logger.fine("Executing module task for "+module.getName());
				
				// Classpath
				List<URL> pathElements = new ArrayList<URL>();
				for (Path path: module.getClassPath()) {
					pathElements.addAll(path.getUrls(factoryConfig.getContextUrl(), factoryConfig.getTokenSource()));
				}
				
				ClassLoader classLoader = factoryConfig.getClassLoader();
				if (classLoader==null) {
					classLoader = getClass().getClassLoader();
				}
				
				if (!pathElements.isEmpty()) {
					logger.fine("Constructing module URL classloader from "+pathElements);
					classLoader = new URLClassLoader(pathElements.toArray(new URL[pathElements.size()]), classLoader);
				}
				
				FactoryConfig moduleFactoryConfig = (FactoryConfig) factoryConfig.clone();
				moduleFactoryConfig.setClassLoader(classLoader);		
				
				final List<ForwardReasoningSessionFactory<Object>> inspectorSetSessionFactories = new ArrayList<ForwardReasoningSessionFactory<Object>>();
				for (InspectorSet inspectorSet: module.getInspectorSet()) {
					inspectorSetSessionFactories.add(createInspectorSetSessionFactory(inspectorSet, moduleFactoryConfig));
				}
				
				try {
					logger.fine("Reviewing module sources: "+module.getSource().size());						
					for (Factory sf: module.getSource()) {
						logger.fine("Reviewing module source");						
						final CompositeObservationSink observationSink = new CompositeObservationSink();
						for (Reporter<Object> r: reporters) {
							observationSink.addSink(r.createObservationSink(module));
						}
						
						FactoryConfig modelConfig = (FactoryConfig) moduleFactoryConfig.clone();
						modelConfig.setContext(new SingletonChainingContext<ObservationSink>(ObservationSink.class, observationSink, modelConfig.getContext()));
						FactoryClosure<Object> sfr = sf.create(modelConfig);
						Object model = sfr.create();
						module.getModel().add(model);
						destroyableSink.addDestroyable(sfr.getDestroyable());						
						
						List<Future<Object>> tasks = new ArrayList<Future<Object>>();
						Iterable<Object> modelIterable = ConvertingService.convert(model, Iterable.class);
						if (modelIterable==null) {
							Warning warning = ReviewFactory.eINSTANCE.createWarning();
							warning.setMessage(model.getClass()+" cannot be converted to Iterable");
							observationSink.consumeObservation(warning);
						} else {
							ExceptionHandler exceptionHandler = new ExceptionHandler() {
								
								@Override
								public void handleException(Exception e) throws Exception {
									Warning warning = ReviewFactory.eINSTANCE.createWarning();
									warning.setMessage("Exception during inspector execution");
									warning.setCause(e);
									observationSink.consumeObservation(warning);
								}
							};
							List<ForwardReasoningSession<Object>> moduleSessions = new ArrayList<ForwardReasoningSession<Object>>();
							for (ForwardReasoningSessionFactory<Object> issf: inspectorSetSessionFactories) {
								logger.fine("Creating module session");
								ForwardReasoningSession<Object> session = issf.createSession(null);
								session.setExceptionHandler(exceptionHandler);
								moduleSessions.add(session); // TODO - session properties.
							}
							try {
								// Collect observations from the model loading
								Iterator<Object> modelIterator = modelIterable.iterator();
								while (modelIterator.hasNext()) {
									Object modelElement = modelIterator.next();
									if (modelElement instanceof LanguageElement) {
										for (Observation observation: ((LanguageElement) modelElement).getObservations()) {
											observationSink.consumeObservation(observation);
										}
									}
								}		

								// Give the model to inspectors.
								modelIterator = modelIterable.iterator();
								while (modelIterator.hasNext()) {
									review(modelIterator, tasks, executor, moduleSessions, inspectorSetSessionFactories, observationSink);							
								}		
								
								// Collect observations from sessions
								for (ForwardReasoningSession<Object> frs: moduleSessions) {
									frs.executeRules();
									for (Object fact: frs.getObjects()) {
										if (fact instanceof Observation) {
											observationSink.consumeObservation((Observation) fact);
										}
									}
								}
								
							} finally {
								for (ForwardReasoningSession<Object> frs: moduleSessions) {
									frs.close();
								}
							}					
						}
						
						// Wait for review tasks to finish
						for (Future<Object> task: tasks) {
							task.get();
						}
						
						// Close sinks
						observationSink.close();
					}
					
					return null;
				} finally {
					for (ForwardReasoningSessionFactory<Object> issf: inspectorSetSessionFactories) {
						if (issf instanceof Destroyable) {
							((Destroyable) issf).destroy();
						}
					}
				}
			}
			
		};
	}

	/**
	 * Reviews model element, possibly in a different thread.
	 * @param modelElement
	 * @param inspectorSetSessionFactories 
	 * @param moduleSessions 
	 * @param counter 
	 * @param executor 
	 * @param observationSinks 
	 * @throws ReasoningException 
	 */
	@SuppressWarnings("unchecked")
	private void review(
			Iterator<Object> modelIterator, 
			final List<Future<Object>> tasks, 
			final Executor executor, 
			List<ForwardReasoningSession<Object>> sessions, 
			final List<ForwardReasoningSessionFactory<Object>> inspectorSetSessionFactories, 
			final ObservationSink observationSink) {
		
		final Object modelElement = modelIterator.next();
		if (modelElement!=null) {
			if (modelIterator instanceof TreeIterator<?> && modelElement.getClass().getAnnotation(Fork.class)!=null) {
				// Reviewing model element in a separate thread
				final Iterable<Object> meIterable = ConvertingService.convert(modelElement, Iterable.class);
				if (meIterable!=null) {
					((TreeIterator<?>) modelIterator).prune();
					Runnable reviewTask = new Runnable() {
	
						@Override
						public void run() {
							try {
								List<ForwardReasoningSession<Object>> meSessions = new ArrayList<ForwardReasoningSession<Object>>();
								for (ForwardReasoningSessionFactory<Object> issf: inspectorSetSessionFactories) {
									meSessions.add(issf.createSession(null)); // TODO - session properties.
								}
								try {
									Iterator<Object> modelIterator = meIterable.iterator();
									while (modelIterator.hasNext()) {
										review(modelIterator, tasks, executor, meSessions, inspectorSetSessionFactories, observationSink);							
									}		
									// Collect observations
									for (ForwardReasoningSession<Object> frs: meSessions) {
										frs.executeRules();
										for (Object fact: frs.getObjects()) {
											if (fact instanceof Observation) {
												observationSink.consumeObservation((Observation) fact);
											}
										}
									}
									
								} finally {
									for (ForwardReasoningSession<Object> frs: meSessions) {
										frs.close();
									}
								}
							} catch (Exception e) {
								Warning warning = ReviewFactory.eINSTANCE.createWarning();
								warning.setMessage("Exception: "+e);
								warning.setCause(e);
								warning.setSource(ConvertingService.convert(modelElement, LanguageElement.class));
							}
						}
						
					};
				
					executor.execute(new NonBlockingFutureTask<Object>(reviewTask, Boolean.TRUE));
					return;
				}
			}
			
			for (ForwardReasoningSession<Object> frs: sessions) {
				try {
					frs.put(modelElement);
				} catch (ReasoningException e) {
					Warning warning = ReviewFactory.eINSTANCE.createWarning();
					warning.setMessage("Exception: "+e);
					warning.setCause(e);
					warning.setSource(ConvertingService.convert(modelElement, LanguageElement.class));
				}
			}
		}		
	}

	protected abstract ForwardReasoningSessionFactory<Object> createInspectorSetSessionFactory(InspectorSet inspectorSet, FactoryConfig moduleFactoryConfig) throws ConfigurationException, ReasoningException;
}
