package com.hammurapi.review.util;

import java.util.logging.Level;
import java.util.logging.Logger;

import com.hammurapi.config.Named;
import com.hammurapi.config.NamedObjectDefinition;
import com.hammurapi.config.ObjectDefinition;
import com.hammurapi.config.bootstrap.ConfigurationException;
import com.hammurapi.config.bootstrap.TokenExpander;
import com.hammurapi.config.bootstrap.TokenExpander.TokenSource;
import com.hammurapi.config.runtime.Component;
import com.hammurapi.config.runtime.ConfigurationContext;
import com.hammurapi.convert.ConvertingService;
import com.hammurapi.reasoning.impl.InferenceMethodNode;
import com.hammurapi.review.Inspector;
import com.hammurapi.review.InspectorSet;
import com.hammurapi.review.LanguageElement;
import com.hammurapi.review.Observation;
import com.hammurapi.review.ReviewFactory;
import com.hammurapi.review.Violation;
import com.hammurapi.review.Warning;

public class InspectorMethodNode extends InferenceMethodNode<Object> implements Component<ObjectDefinition> {
	private static final Logger logger = Logger.getLogger(InspectorMethodNode.class.getName());

	/**
	 * Injects inspector and source into observation.
	 */
	@Override
	protected void onConclusion(Object[] args, Object conclusion) {
		if (inspector!=null && conclusion instanceof Observation) {
			final Observation observation = (Observation) conclusion;
			observation.setReportedBy(inspector);
			if (observation.getSource()==null) {
				for (Object arg: args) {
					LanguageElement src = ConvertingService.convert(arg, LanguageElement.class);
					if (src!=null) {
						observation.setSource(src);
						break;
					}
				}
			}
			
			if (observation instanceof Violation && ((Violation) observation).getMessage()==null && inspector!=null) {
				if (inspector.getMessageTemplate()==null) {
					((Violation) observation).setMessage(inspector.getName()+": "+inspector.getDescription());
				} else {
					TokenSource ts = new TokenSource() {

						@Override
						public String getToken(String name) {
							Violation violation = (Violation) observation;
							Object ret = violation.getProperty(name);
							if (ret!=null) {
								return String.valueOf(ret);
							}
							for (Named property: inspector.getProperty()) {
								if (name.equals(property.getName()) && property instanceof NamedObjectDefinition) {
									return ((NamedObjectDefinition) property).getValue();
								}
							}
							
							return "*** WARNING: Unresolved token "+name+" ***";
						}
						
					};
					
					TokenExpander te = new TokenExpander(ts);
					try {
						((Violation) observation).setMessage(te.expand(inspector.getMessageTemplate()));
					} catch (ConfigurationException e) {
						((Violation) observation).setMessage("Failed to expand inspector message template '"+inspector.getMessageTemplate()+"': "+e);
						logger.log(Level.SEVERE, "Failed to expand inspector message template '"+inspector.getMessageTemplate()+"': "+e, e);
					}
				}
			}
		}
	}
	
	/**
	 * Creates warnings for exceptions in inspector methods.
	 */
	@Override
	protected boolean handleException(Object[] args, Exception e) throws Exception {
		Warning warning = ReviewFactory.eINSTANCE.createWarning();
		warning.setMessage("Exception during inference: "+e);
		warning.setCause(e);
		onConclusion(args, warning);
		knowledgeBase.put(warning);
		return true;
	}

	private Inspector inspector;
	
	@Override
	public void init(ConfigurationContext<ObjectDefinition> context) throws ConfigurationException {
		if (inspectorName!=null) {
			findInspector(context.lookup(InspectorSet.class));
		}
	}

	private boolean findInspector(InspectorSet inspectorSet) {
		for (Inspector i: inspectorSet.getInspectors()) {
			if (inspectorName.equals(i.getName())) {
				inspector = i;
				return true;
			}
		}
		
		for (InspectorSet base: inspectorSet.getBase()) {
			if (findInspector(base)) {
				return true;
			}
		}
		
		return false;
	}

	private String inspectorName;
	
	public void setInspectorName(String inspectorName) {
		this.inspectorName = inspectorName;
	}
}
