package com.hammurapi.review.util;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.LinkedList;

import org.eclipse.emf.ecore.util.EcoreUtil;

import com.hammurapi.config.bootstrap.ConfigurationException;
import com.hammurapi.config.runtime.FactoryConfig;
import com.hammurapi.config.util.ConfigUtil;
import com.hammurapi.flow.Flow;
import com.hammurapi.flow.Node;
import com.hammurapi.reasoning.impl.RuleSetToFlowCompiler;
import com.hammurapi.reasoning.spi.Infer;
import com.hammurapi.reasoning.spi.model.ModelFactory;
import com.hammurapi.reasoning.spi.model.Rule;
import com.hammurapi.reasoning.spi.model.RuleSet;
import com.hammurapi.review.Inspect;
import com.hammurapi.review.Inspector;
import com.hammurapi.review.InspectorCategory;
import com.hammurapi.review.InspectorSet;

/**
 * This class compiles inspector set definition to a flow definition to be used by
 * Hammurapi Rules engine for inferring observations from model elements.
 * @author Pavel Vlasov
 *
 */
public class InspectorSetCompiler  extends RuleSetToFlowCompiler {
	
	/**
	 * Compiles inspector set to flow.
	 * @param source
	 * @param contextClassLoader
	 * @param contextUrl
	 * @param tokens
	 * @param profilePath
	 * @return
	 * @throws ConfigurationException
	 */
	public Flow compile(InspectorSet source, FactoryConfig factoryConfig) throws ConfigurationException {
		return compile(compile(source), factoryConfig);
	}
	
	private RuleSet compile(InspectorSet source) throws ConfigurationException {
		RuleSet ret = ModelFactory.eINSTANCE.createRuleSet();
		InspectorSet inspectorSet = (InspectorSet) EcoreUtil.copy(source);
		
		// Bases
		for (InspectorSet base: inspectorSet.getBase()) {
			ret.getBase().add(compile(base));
		}
		
		ConfigUtil.pump(inspectorSet, ret);

		// Inspectors
		for (Inspector inspector: inspectorSet.getInspectors()) {
			if (inspector.isEnabled() && inspector.isRuntime()) {
				Rule rule = ModelFactory.eINSTANCE.createRule();
				ret.getRule().add(rule);
				ConfigUtil.pump(inspector, rule);
				ConfigUtil.addScalarProperty(rule, "severity", inspector.getSeverity(), false);
				LinkedList<String> cl = new LinkedList<String>();
				for (InspectorCategory c = inspector.getCategory(); c!=null; c = c.getParentCategory()) {
					cl.addFirst(c.getName());
				}
				for (String cn: cl) {
					ConfigUtil.addScalarProperty(rule, "category", cn, false);
				}
			}
		}
		
		return ret;
	}
	
	@Override
	protected Node createInferenceMethodNode(
			RuleSet ruleSet, 
			Rule rule,
			Method method, Flow owner) {
		Node ret = createNode(owner, InspectorMethodNode.class);
		ConfigUtil.addScalarProperty(ret, "inspectorName", rule.getName(), true);
		return ret;
	}
	
	@Override
	protected Infer getInferAnnotation(Method method) {
		final Inspect inspect = method.getAnnotation(Inspect.class);
		if (inspect==null) {
			return null;
		}
		
		return new Infer() {

			@Override
			public int priority() {
				return 0;
			}

			@Override
			public Class<?>[] value() {
				return inspect.value();
			}

			@Override
			public Class<? extends Annotation> annotationType() {
				return Infer.class;
			}
			
		};
	}

}
